/* This application is for the following board combination:
 * TS-8100, TS-4740, TS-ADC24.
 * The purpose is to leverage the Spartan 6 and LM32 to maximize bandwidth 
 * between the TS-4740 CPU and the TS-ADC24.
 *
 * There is an endianness conflict between the LM32 (big endian) and the rest
 * of the hardware.  Both the ARM9 and the peripherals are little endian.
 * For performance reasons, we must work with 16 bit values, so this must be
 * accounted for.  The PEEK16 and POKE16 macros are for accessing 16 bit
 * little endian registers in LM32 code.
 *
 * In order for the LM32 and ARM9 to pass 16 bit data back and forth without
 * doing an explicit endianness conversion, we treat regions of LM32 memory
 * as hardware registers.  Here are the steps to doing this:
 * - declare an unsigned short (so that the memory is allocated)
 * - cast the address of that variable to an unsigned int
 * - add 0x10000
 * - use PEEK16/POKE16 on the resulting value.
 * The 0x10000 adjustment is necessary because PEEK16/POKE16 of values below
 * 0x10000 go to hardware registers, whereas 0x10000 to 0x1ffff goes to memory.
 */

#include "ts4740.h"

int main(void)
{
	unsigned short buffer_sh[0x2000];
	unsigned short tail_sh;
	register unsigned int tail, ready;
	unsigned int head, fullness;
	volatile unsigned int buffer, tail_p;
	volatile register unsigned short *dat;
	register unsigned int x;
	//register unsigned int a, b;
	register unsigned int *p, *q;
	// buffer_sh and tail_sh are only used to allocate space for buffer
	// and tail_p

	ack_startup();
	outstr("\nTS-ADC24 buffering app.\n\n");
	
	// shared pointer setup
	buffer = (unsigned int)buffer_sh + 0x10000;
	for (tail=0; tail != 0x4000; tail+=2) POKE16(buffer + tail, 0);
	tail_p = (unsigned int)&tail_sh + 0x10000;
	outstr("\nbuffer=0x");
	outhex((unsigned int)buffer);
	outstr("\ntail_p=0x");
	outhex((unsigned int)tail_p);

	// baseboard access:
	POKE16(MUXBUS_CONFIG, 0x321);
	outstr("\nBaseboard ID: 0x");
	outhex(PEEK16(MUXBUS16));
	outstr("\nPeripheral ID: ");
	outhex(PEEK16(MUXBUS16 + 0x100));
	outstr("\n\n");

#define IRQ_ON 0x41
#define IRQ_OFF 0x4000
	// begin TS-ADC24 monitoring
	tail = 0;
	dat = (unsigned short *)(BASE16 + MUXBUS16 + 0x10a);
	p = (unsigned int *)(BASE32 + buffer);
	POKE32(0x4e88, buffer); // sane start condition for head
	//POKE16(MUXBUS16 + 0x102, PEEK16(MUXBUS16 + 0x102) | 1); // XXX
	while (1) {
		POKE16(tail_p, tail);
		ready = PEEK16(MUXBUS16 + 0x108) >> 6;
		head = (PEEK32(0x4e88) - buffer) >> 1;
		if (head > tail) fullness = tail + 0x2000 - head;
		else fullness = tail - head;
		if (fullness & 0x1000) POKE16(MUXBUS16 + 0x108, IRQ_ON);
		else POKE16(MUXBUS16 + 0x108, IRQ_OFF);
		if (ready + tail > 0x2000) ready = 0x2000 - tail;
		if (fullness + ready >= 0x2000) ready = 0x2000 - fullness - 1;
		ready = ready & ~0x3f; // 64 word chunks
		q = (p + (ready>>1));
		/*
		outstr("tail=0x");
		outhex(tail);
		outstr(" head=0x");
		outhex(head);
		outstr(" fullness=0x");
		outhex(fullness);
		outstr(" ready=0x");
		outhex(ready);
		outstr("\n");
		*/
		while (p != q) {
			/*
			a = *dat << 20;
			b = *dat & 0xfff;
			a |= b << 8;
			b = *dat & 0xfff;
			a |= b >> 4;
			*p++ = a;
			a = b << 28;
			b = *dat & 0xfff;
			a |= b << 16;
			b = *dat & 0xfff;
			a |= b;
			b = *dat & 0xfff;
			a |= b >> 8;
			*p++ = a;
			a = b << 24;
			b = *dat & 0xfff;
			a |= b << 12;
			b = *dat & 0xfff;
			a |= b;
			*p++ = a;
			// up to this point I wrote 12 bytes, 8 samples
			a = *dat << 20;
			b = *dat & 0xfff;
			a |= b << 8;
			b = *dat & 0xfff;
			a |= b >> 4;
			*p++ = a;
			a = b << 28;
			b = *dat & 0xfff;
			a |= b << 16;
			b = *dat & 0xfff;
			a |= b;
			b = *dat & 0xfff;
			a |= b >> 8;
			*p++ = a;
			a = b << 24;
			b = *dat & 0xfff;
			a |= b << 12;
			b = *dat & 0xfff;
			a |= b;
			*p++ = a;
			// up to this point I wrote 24 bytes, 16 samples
			a = *dat; a <<= 16; a |= *dat; *p++ = a;
			a = *dat; a <<= 16; a |= *dat; *p++ = a;
			// Now I wrote 32 bytes, 20 samples.
			*/

			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
			x = *dat; x <<= 16; x |= *dat; *p++ = x;
		}
		tail += ready;
		if (tail == 0x2000) {
			tail = 0;
			p = (unsigned int *)(BASE32 + buffer);
		}
	}

}

